#define COUNT_TESTS              1      /* count tests for ... */
#define SOURCE_CMDLINE_PARAM_NOM 1      /* number of commandline parameter containing source filename for evaluating */
#define INFILE_CMDLINE_PARAM_NOM 2      /* number of commandline parameter containing input data filename */

char * msg[] = {
  "%f\n",
  "0\n",
  "At row %d: Unknown command %s",
  "At row %d: Missing argument(s) for command %s",
  "At row %d: Bad variable name (%s)",
  "At row %d: Undefined variable %s",
  "At row %d: Missing second argument for command %s",
  "At row %d: To many arguments for command %s",
  "At row %d: Try to read data after end of input (%s)",
  "At row %d: Bad variable name or numeric constant (%s)",
  "At row %d: Redefine variable (%s)",
  "=== ERRORS ========================\n",
  "=== WARNINGS ======================\n",
  "=== TESTS RESULTS =================\n",
  "\n\n",
  "=== Outputs by tests ==============\n",
  "Randomly selected %d%% of the tests achieve %f points\n",
  "Can't open source file (or file not exist)\n",
  "'Empty' program (not to do)\n",
  "Can't open input datafile (or file not exist)\n",
  "Error reading input datafile\n"

};

#include<fstream>
#include <iostream>
#include <cstring>
#include <vector>
#include <map>
#include <list>
#include <queue>
#include <climits>
#include <cstdlib>
#include <ctime>
#include <cstdio>


using namespace std;

struct TCmd{short cmd, a, b;};
typedef vector <long long> TData;
typedef vector <TCmd> TCode;
typedef map <string, short> TTokens;
typedef list <string> TStrings;
typedef queue <long long> TStreamLongLong;

TStreamLongLong stdInput, stdOutput;
TData Data, stdSols, Output[COUNT_TESTS];
TCode Code;
TTokens Vars, Commands;
TStrings Errors , Warnings;
double Results[COUNT_TESTS];

bool execute(int);
void print_err(int , TStrings& );
bool init_checker(char*);
void print_result();
bool read_input_datafile(char *filename, TStreamLongLong& ins);
void print_help(char *pgmname);

int main(int argc, char * argv[]){
  if (argc==0)
  if (argc<=SOURCE_CMDLINE_PARAM_NOM) Errors.push_back(msg[17]);
  if (argc<=INFILE_CMDLINE_PARAM_NOM) Errors.push_back(msg[19]);
  if (Errors.size()) {
    print_help(argv[0]);
    return 0;
  }
//  cout << argc << endl;
//  cout << argv[INFILE_CMDLINE_PARAM_NOM] << endl;
  if (!read_input_datafile(argv[INFILE_CMDLINE_PARAM_NOM],stdInput) or !init_checker(argv[SOURCE_CMDLINE_PARAM_NOM]))
    print_err(11,Errors);
  else {
    for(int i=0;i<COUNT_TESTS;i++) { execute(i); }
    print_result();
  }
  return 0;
}

int _count_cmds[8];

void f_DEF(TData& d, short a, short b=0){
  d[a]=rand();
}
void f_ADD(TData& d, short a, short b){
  d[a]+=d[b];
  if (d[a]<INT_MIN or d[a]>INT_MAX) d[a]=rand();
}
void f_SUB(TData& d, short a, short b){
  d[a]-=d[b];
  if (d[a]<INT_MIN or d[a]>INT_MAX) d[a]=rand();
}
void f_MUL(TData& d, short a, short b){
  d[a]*=d[b];
  if (d[a]<INT_MIN or d[a]>INT_MAX) d[a]=rand();
}
void f_DIV(TData& d, short a, short b){
  if (d[b]) d[a]/=d[b];
  else d[a]=rand();
}
void f_GET(TData& d, short a, short b){
  if (stdInput.size()) {
    d[a]=stdInput.front();
    stdInput.pop();
  }
}
void f_PUT(TData& d, short a, short b){
  stdOutput.push(d[a]);
}
void f_NOP(TData& d, short a, short b){}

typedef void (*TFunc) (TData &, short, short);
TFunc F[]={f_NOP, f_DEF, f_GET, f_PUT, f_ADD, f_SUB, f_MUL, f_DIV};

void add_err(int row, int mess_nom, char *a){
  char buff[256];
  sprintf(buff,msg[mess_nom],row,a);
  Errors.push_back(buff);
}
bool file_exist(char* fn){
  bool res = false;
  ifstream f;
  f.open(fn);
  if (f) {
    res = true;
    f.close();
  }
  return res;
}
int check_cmd(char *c){
  int rez=0;
  char *p=c;
  for(;*c;c++)
    if (isalpha(*c)) *c=toupper(*c);
    else rez--;
  if (rez==0) {
    try {
      rez = Commands.at(p);
    }
    catch (...){return 0;}
  }
  return rez;
}
bool check_token(char *c){
  bool rez=true;
  for (;*c;c++)
    if (isalpha(*c)) *c=toupper(*c);
    else rez=false;
  return rez;
}
bool token_exist(char *c, short& pos){
  try{
    short i = Vars.at(c);
    pos=i;
    return true;
  }
  catch (...){return false; }
}
bool token_is_num(char *c,short& pos){
  char *p=c;
  long long j;
  if (c == NULL) return false;
  else {
    if (*c=='-' or *c=='+') c++;
    for(;*c;c++) if (!isdigit(*c)) return false;
    if (!token_exist(p,pos)) {
      if (sscanf(p,"%lld",&j)==0) return false;
      pos=Data.size();
      Data.push_back(j);
      Vars[p]=pos;
    }
  }
  return true;
}

bool read_input_datafile(char *filename, TStreamLongLong& ins){
  FILE *f;
  int i;
  bool res=true;

  while (!ins.empty()) ins.pop();

  if (!file_exist(filename)) {Errors.push_back(msg[19]);res=false;}
  else {
    f=fopen(filename,"r");
    //cout << "in file open\n";
    if (f) {
      while (!feof(f) and fscanf(f," %d",&i)==1) ins.push(i);
      if (!feof(f) and ins.size()==0) {Errors.push_back(msg[20]);res=false;}
      }
    else {Errors.push_back(msg[19]);res=false;}
    if (f) fclose(f);
  }
  return res;
}

bool init_checker(char *source_name){
  char s[2048], *cmd, *farg, *sarg, *targ, *p;
  int i,j;
  TCmd _cmd;
  FILE *f;

  //init rand gen.
  srand (time(NULL));

  //init Commands
  char *cmds[]={"DEF", "GET", "PUT", "ADD", "SUB", "MUL", "DIV"};
  for (i=0;i<7;i++) Commands[cmds[i]]=i+1;

  //read pgm source and compile
  if (!file_exist(source_name)){
    Errors.push_back(msg[17]);
    return false;
  }
  f=fopen(source_name,"r");
  if (f)
    for (i=1;!feof(f);i++){
      s[0]='\0';
      if (!fgets(s,2047,f)) continue;
      p = strpbrk(s,";\n\r");
      if (p) *p='\0';
      if (s[0]=='\0') continue;
      cmd=strtok(s," \t");
      farg=strtok(NULL," \t");
      sarg=strtok(NULL," \t");
      targ=strtok(NULL," \t");

      //check command
      if (cmd == NULL) continue;
      _cmd.cmd=check_cmd(cmd);
      if (_cmd.cmd<=0) {add_err(i,2,cmd); continue;}

      // check first argument
      if (farg==NULL) { add_err(i,3,farg); continue; }
      if (!check_token(farg))add_err(i,4,farg);
      else if (token_exist(farg,_cmd.a)){ if (_cmd.cmd==1) add_err(i,10,farg); }
      else if (_cmd.cmd>1) add_err(i,5,farg);
      else {
        _cmd.a=Data.size();
        Vars[farg]=_cmd.a;
        Data.push_back(0);
      }

      // check second argument
      if (sarg==NULL) { if (_cmd.cmd>3) {add_err(i,6,cmd); continue;}}
      else if (_cmd.cmd<4) { add_err(i,7,cmd); continue; }
      else if (check_token(sarg)) {if (!token_exist(sarg,_cmd.b)) add_err(i,5,sarg);}
      else if (!token_is_num(sarg,_cmd.b)) add_err(i,9,sarg);

      // check for extra terms
      if (targ != NULL) {add_err(i,7,cmd); continue; }

      Code.push_back(_cmd);
      _count_cmds[ _cmd.cmd ]++;
      if (_cmd.cmd==2 and _count_cmds[2]>stdInput.size()) add_err(i,8,"check input data stream for errors!");
    }
  else Errors.push_back(msg[17]);
  if (f) fclose(f);

  if (Code.size()==0) Errors.push_back(msg[18]);
  return Errors.size()==0;
}

void AssignInput(TStreamLongLong& inp, char* s, TData& sol){
  long long sums[4]={0,0,0,0};
  char q[512], *p;
  int  w,i,j,k;
  while (!inp.empty()) inp.pop();
  strcpy(q,s);
  p=strtok(q," \t");
  for (j=0;p;j++){
    k=sscanf(p," %d", &i);
    if (j&1){
      switch (i){
        case 0xffff  : sums[0]+=w;break;
        case 0xff00ff: sums[1]+=w;break;
        case 0xffff00: sums[2]+=w;break;
        default: sums[3]+=w;
      }
      inp.push(w);
      inp.push(i);
      }
    else w=i;
    p=strtok(NULL," \t");
  }
  sol.clear();
  for (i=0;i<4;i++) sol.push_back(sums[i]);
}
bool execute(int test){
  int i,j=Code.size();
  ///AssignInput(stdInput, inp_data[test], stdSols);
  while (!stdOutput.empty()) stdOutput.pop();
  for (i=0;i<j; i++) F[Code[i].cmd](Data, Code[i].a, Code[i].b );
}
void print_err(int nom, TStrings& s){
  printf(msg[nom]);
  for (TStrings::iterator i=s.begin(); i != s.end(); ++i ) printf("%s\n",(*i).c_str());
}
void print_result(){
  int i,k;
  double j;
  long long d;

  if (Warnings.size()) print_err(12,Warnings);
  if (Errors.size()) print_err(11,Errors);
  printf("\nPROGRAM\'s OUTPUT:\n");
  for (i=0;i<COUNT_TESTS;i++) {
    ///fprintf(stderr,"%2d:",i+1);
    for (j=0,k=stdOutput.size();j<k;j++) {
      printf("%lld\n",stdOutput.front());
      stdOutput.pop();
    }
    ///fprintf(stderr,"\n");
  }

}
void print_help(char *pgmname){
  printf("USE:\n\n    %s sourcefilename inputdatastream\n\n",pgmname);
  printf("Example:\n%s my_pgm.txt my_data_for_this_pgm.txt\n\n",pgmname);
}
